//
//  HSNavigationController.swift
//  HSNavigationController
//
//  Created by zwb on 17/1/6.
//  Copyright © 2017年 hengsu technology. All rights reserved.
//

import UIKit

public typealias HSNavigationControllerCompletionBlock = (()->Void)?

open class HSNavigationController: UIViewController , UIGestureRecognizerDelegate{

    /// All viewController add and exist.
    open var viewControllers:[UIViewController]!
    
    /// the default animation time when push or pop. And you can setting.
    private static let animationDuration:TimeInterval=0.5
    /// when the animationing, the max alpha for the maskView.
    /// if you needn't maskView, change it to 0.0
    private static let maxBlackMaskAlpha:CGFloat=0.8/*0.0*/
    /// when the animationing, the previousViewController's view transform scale ration.
    /// if you needn't transform, change it to 1.0
    private static let zoomRatio:CGFloat=0.9/*1.0*/
    /// when the animationing, the previousViewController's view transform opacity.
    /// if you needn't layer opacity, change it to 1.0
    private static let shadowOpacity:Float=0.8/*1.0*/
    /// when the animationing, the previousViewController's view transform radius.
    /// if you needn't layer radius, change it to 0.0
    private static let shadowRadius:CGFloat=8.0/*0.0*/
    
    private var gestures:[UIPanGestureRecognizer]!
    private var blackMask:UIView!
    private var animationing:Bool!
    private var percentageOffsetFromLeft:CGFloat!
    
    private enum PanDirection : Int
    {
        case none
        case left
        case right
    }
    
    ///  Init
    ///  initWithRootViewController
    public convenience init(withRootViewController rootViewController:UIViewController)
    {
        
        self.init()
        
        viewControllers=[rootViewController]
    }
    
    ///  Push
    ///  push to the ‘viewController’ and animationing
    ///  after stopAnimation add panGesture to the viewController's view
    ///
    /// - parameter viewController  : Will push controller
    /// - parameter completion      : The compente Block
    /// - parameter responseGesture : Add panGesture to the controller's view
    ///
    public func pushViewController(_ viewController:UIViewController, Completion completion: HSNavigationControllerCompletionBlock?=nil, ResponseGesture responseGesture:Bool=true)
    {
        
        animationing=true
        
        viewController.view.layer.shadowColor=UIColor.black.cgColor
        viewController.view.layer.shadowOpacity=HSNavigationController.shadowOpacity
        viewController.view.layer.shadowRadius=HSNavigationController.shadowRadius
        
        viewController.view.frame=view.bounds.offsetBy(dx: view.bounds.size.width, dy: 0)
        viewController.view.autoresizingMask=[.flexibleWidth,.flexibleHeight]
        blackMask.alpha=0
        viewController.willMove(toParentViewController: self)
        addChildViewController(viewController)
        
        view.bringSubview(toFront: blackMask)
        view.addSubview(viewController.view)
        
        UIView.animate(withDuration: HSNavigationController.animationDuration, delay: 0, options: .curveEaseInOut, animations: {
            
            let transform=CGAffineTransform.identity
            self.currentViewController()!.view.transform=transform.scaledBy(x: HSNavigationController.zoomRatio,
                                                                            y: HSNavigationController.zoomRatio)
            viewController.view.frame=self.view.bounds
            self.blackMask.alpha=HSNavigationController.maxBlackMaskAlpha
            
        }) { (finished) in
            
            if finished {
                
                self.viewControllers.append(viewController)
                viewController.didMove(toParentViewController: self)
                self.animationing=false
                self.gestures=[UIPanGestureRecognizer]()
                
                if responseGesture { self.addPanGestureToView(viewController.view) }
                
                if let complete=completion {  complete!()  }
            }
        }
    }
    
    ///  Po
    ///  pop to the previousViewController and animationing
    ///  after stopAnimation remove controller from parentController
    ///
    /// - parameter completion      : The compente Block
    ///
    public func popViewControllerCompletion(_ completion:HSNavigationControllerCompletionBlock?=nil)
    {
        if viewControllers.count<2 { fatalError("HSNavigationController's viewControllers the count less than 2.") }
        animationing=true
        
        let currentVc=currentViewController()
        let previousVc=previousViewController()
        previousVc!.viewWillAppear(false)
        
        currentVc!.view.layer.shadowColor=UIColor.black.cgColor
        currentVc!.view.layer.shadowOpacity=HSNavigationController.shadowOpacity
        currentVc!.view.layer.shadowRadius=HSNavigationController.shadowRadius
        
        UIView.animate(withDuration: HSNavigationController.animationDuration, delay: 0, options: .curveEaseInOut, animations: { 
            
            currentVc!.view.frame=self.view.bounds.offsetBy(dx: self.view.bounds.size.width, dy: 0)
            let transform=CGAffineTransform.identity
            previousVc!.view.transform=transform.scaledBy(x: 1.0, y: 1.0)
            previousVc!.view.frame=self.view.bounds
            self.blackMask.alpha=0
            
        }) { (finished) in
            
            if finished {
                
                currentVc!.view.removeFromSuperview()
                currentVc!.willMove(toParentViewController: nil)
                self.view.bringSubview(toFront: previousVc!.view)
                currentVc!.removeFromParentViewController()
                currentVc!.didMove(toParentViewController: nil)
                self.viewControllers.removeLast()
                
                self.animationing=false
                previousVc!.viewDidAppear(true)
                
                if let complente=completion { complente!() }
                
                HSLog("The current viewControllers count is \(self.viewControllers.count)")
            }
        }
    }
    
    ///  Pop
    ///  pop to the ‘viewController’ and animationing
    ///  after stopAnimation remove between the currentController with viewController from parentController
    ///
    /// - parameter viewController  : Will pop to the controller
    ///
    public func popViewController(toViewController viewController:UIViewController, completion Completion:HSNavigationControllerCompletionBlock?=nil)
    {
        guard let vcs=viewControllers else {
            fatalError("HSNavigationController's viewControllers is nil")
        }
        guard let index=vcs.index(of: viewController) else{
            fatalError("HSNavigationController's viewControllers not contains \(viewController)")
        }
        if vcs.count<2{
            fatalError("HSNavigationController's viewControllers cout less than 2")
        }
        var needRemoveViewController:UIViewController!
        
        for i in stride(from: vcs.count-2, to: index, by: -1) {
            
            needRemoveViewController=vcs[i]
            needRemoveViewController.view.alpha=0
            needRemoveViewController.removeFromParentViewController()
            viewControllers.remove(at: i)
        }
        
        popViewControllerCompletion(Completion)
    }
    
    ///  Pop
    ///  pop to the rootViewController and animationing
    ///  after stopAnimation remove viewControllers from parentController outside rootViewController
    ///
    /// - parameter viewController  : Will pop to the controller
    ///
    public func popToRootViewController(_ completion:HSNavigationControllerCompletionBlock?=nil)
    {
        guard let vcs=viewControllers else {
            fatalError("HSNavigationController's viewControllers is nil")
        }
        guard let firstVc=vcs.first else {
            fatalError("HSNavigationController's viewControllers count is 0")
        }
        popViewController(toViewController: firstVc, completion: completion)
    }
    
    open override func loadView()
    {
        
        super.loadView()
        
        let viewRect=viewBoundsWithOrientation(UIApplication.shared.statusBarOrientation)
        
        // add ChildViewController
        guard let vcs=viewControllers else {
            fatalError("HSNavigationController's viewControllers is nil")
        }
        guard let rootViewController=vcs.first else {
            fatalError("HSNavigationController's viewControllers the first is nil")
        }
        rootViewController.willMove(toParentViewController: self)
        addChildViewController(rootViewController)
        
        // add ChildViewController's view
        let rootView=rootViewController.view!
        rootView.autoresizingMask=[.flexibleWidth,.flexibleHeight]
        rootView.frame=viewRect
        view.addSubview(rootView)
        rootViewController.didMove(toParentViewController: self)
        
        // add ChildViewController's view maskView
        let bkMask=UIView(frame: viewRect)
        bkMask.backgroundColor=UIColor.black
        bkMask.alpha=0
        bkMask.autoresizingMask=[.flexibleHeight,.flexibleWidth]
        view.insertSubview(bkMask, at: 0)
        blackMask=bkMask
        
        view.autoresizingMask=[.flexibleWidth,.flexibleHeight]
    }
    
    // ---- MARK: UIPanGestureRecognizer
    /// add PanGesture
    private func addPanGestureToView(_ view:UIView)
    {
        
        let panGesture=UIPanGestureRecognizer(target: self,
                                              action: #selector(panGestureResponse(_:)))
        panGesture.delegate=self
        view.addGestureRecognizer(panGesture)
        gestures.append(panGesture)
    }
    
    /// Response UIPanGestureRecognizer
    func panGestureResponse(_ panGesture:UIPanGestureRecognizer)
    {
        
        if animationing! {return}
        
        let currentPoint=panGesture.translation(in: view)
        let x=currentPoint.x
        
        var panDirection=PanDirection.none
        let vel=panGesture.velocity(in: view)
        
        if vel.x>0{
            panDirection = .right
        }else{
            panDirection = .left
        }
        
        var offset:CGFloat=0
        let vc=currentViewController()
        offset=vc!.view.frame.size.width-x
        vc!.view.layer.shadowColor=UIColor.black.cgColor
        vc!.view.layer.shadowOpacity=HSNavigationController.shadowOpacity
        vc!.view.layer.shadowRadius=HSNavigationController.shadowRadius
        
        percentageOffsetFromLeft=offset/viewBoundsWithOrientation(UIApplication.shared.statusBarOrientation).size.width
        vc!.view.frame=getSlidingRectWithPercentageOffset(percentageOffsetFromLeft,
                                                          Orientation: UIApplication.shared.statusBarOrientation)
        transformAtPercentage(percentageOffsetFromLeft)
        
        if panGesture.state == .ended || panGesture.state == .cancelled{
            
            if fabs(vel.x)>100{
                
                completeSlidingAnimationWithDirection(panDirection)
            }else{
                
                completeSlidingAnimationWithOffset(offset)
            }
        }
    }
    
    // ---- MARK : UIGestureRecognizerDelegate
    ///
    
    // MARK: Private Methods
    private func viewBoundsWithOrientation(_ orientation:UIInterfaceOrientation) -> CGRect
    {
        
        var bounds=UIScreen.main.bounds
        
        // if UIApplicationStatusBar is hidden , the bounds is Screen's bounds
        if UIApplication.shared.isStatusBarHidden { return bounds }
            
        else if UIInterfaceOrientationIsLandscape(orientation){
            
            let width=bounds.size.width
            bounds.size.width=bounds.size.height
            bounds.size.height=width
            return bounds
            
        }else{
            
            return bounds
        }
    }
    
    // Get the top viewcontroller
    private func currentViewController() -> UIViewController?
    {
    
        guard let vcs=viewControllers else {
            fatalError("HSNavigationController's viewControllers is nil")
        }
        if vcs.count>0 { return vcs.last }
        return nil
    }
    
    // Get the previous Viewcontroller
    private func previousViewController() -> UIViewController?
    {
        
        guard let vcs=viewControllers else {
            fatalError("HSNavigationController's viewControllers is nil")
        }
        if vcs.count>1 { return vcs[vcs.count-2] }
        return nil
    }
    
    
    /// Complete this Animation when called on finger up when the panGesture velocity
    /// the distance greater 100
    private func completeSlidingAnimationWithDirection(_ direction:PanDirection)
    {
        
        if direction == .right{
            
            popViewControllerCompletion()
        }else{
            
            rollBackViewController()
        }
    }
    
    /// Complete this Animation when called on finger up when the panGesture velocity
    /// the distance less than 100
    private func completeSlidingAnimationWithOffset(_ offset:CGFloat)
    {
        
        if offset<viewBoundsWithOrientation(UIApplication.shared.statusBarOrientation).size.width*0.5{
            
            popViewControllerCompletion()
        }else{
            
            rollBackViewController()
        }
    }
    
    private func rollBackViewController()
    {
        
        animationing=true
        
        let vc=currentViewController()
        let preVc=previousViewController()
        let rect=CGRect(x: 0,
                        y: 0,
                        width: vc!.view.frame.size.width,
                        height: vc!.view.frame.size.height)
        
        UIView.animate(withDuration: 0.3, delay: 0, options: .curveEaseInOut, animations: { 
            
            let transform=CGAffineTransform.identity
            preVc!.view.transform=transform.scaledBy(x: HSNavigationController.zoomRatio,
                                                     y: HSNavigationController.zoomRatio)
            vc!.view.frame=rect
            self.blackMask.alpha=HSNavigationController.maxBlackMaskAlpha
            
        }) { (finished) in
            
            if finished { self.animationing=false }
        }
    }
    
    private func getSlidingRectWithPercentageOffset(_ percentage:CGFloat, Orientation orientation:UIInterfaceOrientation)-> CGRect
    {
        
        let viewRect=viewBoundsWithOrientation(UIApplication.shared.statusBarOrientation)
        var rectToRetrun=CGRect.zero
        rectToRetrun.size=viewRect.size
        let x=(1-percentage)*viewRect.size.width>0 ? (1-percentage)*viewRect.size.width : 0
        rectToRetrun.origin=CGPoint(x: x, y: 0)
        
        return rectToRetrun
    }
    
    private func transformAtPercentage(_ percentage: CGFloat)
    {
        
        let transform=CGAffineTransform.identity
        let newTransformValue=1-percentage*(1-HSNavigationController.zoomRatio)
        let newAlphaValue=percentage*HSNavigationController.maxBlackMaskAlpha
        previousViewController()!.view.transform=transform.scaledBy(x: newTransformValue,
                                                                    y: newTransformValue)
        blackMask.alpha=newAlphaValue
    }
    
}


// -----  MARK: Extension UIViewController
extension UIViewController {
    
    public dynamic var hsNavigationController : HSNavigationController!
    {
        get
        {
            
            var responder=next
            
            while responder != nil {
                
                if responder!.isKind(of: HSNavigationController.classForCoder())
                {
                    return responder as! HSNavigationController
                }
                
                responder=responder?.next
            }
            
            return nil
        }
    }
    
}

// -----  MARK: DEBUG Log
public func HSLog<T>(_ message:T, file File:NSString=#file, method Method:String=#function, line Line:Int=#line) {
    #if DEBUG
        print("\(File.lastPathComponent)[\(Line)], \(Method): \(message)")
    #endif
}
